import django_jalali.db.models as jmodels

from django.db import models
from rest_framework.exceptions import AuthenticationFailed

from django.contrib.auth import get_user_model


NUM_ALLOWED_TOKENS_FOR_USER = 3


User = get_user_model()


class TokenLookUpIDManager(models.Manager):

    def create_token(self, user, ip, r_type, browser, os, device):
        user_tokens_queryset = self.filter(user=user)

        # First check if this user has exceeded maximum allowed tokens or not!
        if user_tokens_queryset.count() >= NUM_ALLOWED_TOKENS_FOR_USER:
            # If this happened. remove latest token and create a new one

            if not user.is_staff:
                last_tokens = user_tokens_queryset.order_by('-created_at')[:NUM_ALLOWED_TOKENS_FOR_USER-1]

                for token in user_tokens_queryset.exclude(id__in=last_tokens):
                    token.delete()

        # Now create a new user and assign an ID to it.
        return self.create(user=user, ip=ip, os=os[:30], browser=browser[:30],
                           device=device[:30], r_type=r_type[:30])

    def check_token(self, user_id, lookup_id):
        if self.filter(user__id=user_id, id=lookup_id).exists():
            return True
        raise AuthenticationFailed('توکن معتبر نیست.')


class TokenLookUpID(models.Model):
    """
    This is the main model to control tokens generated with jwt.\
    because of jwt, we don't have control on tokens after expire or user revoke.\
    So we map each created jwt token to an instance of this model.\

    For example if we define that only 3 active tokens\
    should be created for each user. we do as follow:
        - First create a jwt token for user, but first create\
        a lookupID for it and store that id in payload of jwt.
        - Then each time a request came with jwt, check payload\
        and see the provided lookup_id is in list of\
        lookupIDs for that user (max 3 is defined for this case)
        - each time user revoked this token or has been expired,\
        then remove lookup_id or assign a new one on jwt refresh.
    """
    # user_id = models.IntegerField(db_index=True)
    user = models.ForeignKey(User, db_index=True, on_delete=models.CASCADE)

    ip = models.CharField(max_length=20, null=True, blank=True)
    os = models.CharField(max_length=300, null=True, blank=True)
    r_type = models.CharField(max_length=20, null=True, blank=True)
    device = models.CharField(max_length=20, null=True, blank=True)
    browser = models.CharField(max_length=30, null=True, blank=True)

    created_at = jmodels.jDateTimeField(auto_now_add=True)
    updated_at = jmodels.jDateTimeField(auto_now=True)

    objects = TokenLookUpIDManager()

    class Meta:
        db_table = 'token_manager_lookup_id'

    def __str__(self):
        return '{}, {}'.format(self.user.get_full_name(), self.id)



from django.conf.urls import url, include
from rest_framework import routers

from .views import *

router = routers.DefaultRouter()

router.register(r'manage', JSONWebTokenView, basename='manage_tokens')


urlpatterns = [
    url(r'^get/', obtain_jwt_token, name='obtain_jwt_token'),
    url(r'^verify/', verify_jwt_token, name='verify_jwt_token'),
    url(r'^logout/', logout_view, name='logout_view'),

    url(r'^', include(router.urls)),

]



import jwt
from django.contrib.auth import authenticate, get_user_model
from django.utils.translation import ugettext as _
from rest_captcha.serializers import RestCaptchaSerializer
from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from token_manager.utils import get_client_ip, get_lookup_id_from_request, \
    fetch_request_extra_info
from .compat import Serializer, get_username_field, PasswordField
from .models import TokenLookUpID
from .settings import api_settings

User = get_user_model()
jwt_payload_handler = api_settings.JWT_PAYLOAD_HANDLER
jwt_encode_handler = api_settings.JWT_ENCODE_HANDLER
jwt_decode_handler = api_settings.JWT_DECODE_HANDLER
jwt_get_username_from_payload = api_settings.JWT_PAYLOAD_GET_USERNAME_HANDLER


class JSONWebTokenSerializer(Serializer):
    """
    Serializer class used to validate a username and password.

    'username' is identified by the custom UserModel.USERNAME_FIELD.

    Returns a JSON Web Token that can be used to authenticate later calls.
    """

    def __init__(self, *args, **kwargs):
        """
        Dynamically add the USERNAME_FIELD to self.fields.
        """
        super(JSONWebTokenSerializer, self).__init__(*args, **kwargs)

        self.fields[self.username_field] = serializers.CharField()
        self.fields['password'] = PasswordField(write_only=True)

    @property
    def username_field(self):
        return get_username_field()

    def validate(self, attrs):
        credentials = {
            self.username_field: attrs.get(self.username_field),
            'password': attrs.get('password')
        }

        if all(credentials.values()):
            user = authenticate(**credentials)

            if user:
                if not user.is_active:
                    # msg = 'حساب کاربری فعال نیست.'
                    msg = 'Account is not active.'
                    raise serializers.ValidationError(msg)

                request = self.context.get('request', None)

                ip, r_type, browser, os, device = fetch_request_extra_info(request)

                payload, lookup_id = jwt_payload_handler(user, ip, r_type, browser, os, device)

                return {
                    'token': jwt_encode_handler(payload),
                    'user': user
                }
            else:
                # msg = 'امکان ورود با اطلاعات ارائه شده وجود ندارد.'
                msg = 'Unable to login with provided credentials.'
                raise serializers.ValidationError(msg)
        else:
            msg = _('Must include "{username_field}" and "password".')
            msg = msg.format(username_field=self.username_field)
            raise serializers.ValidationError(msg)


# class JSONWebTokenWithCaptchaSerializer(RestCaptchaSerializer, JSONWebTokenSerializer):
class JSONWebTokenWithCaptchaSerializer(JSONWebTokenSerializer):
    pass


class VerificationBaseSerializer(Serializer):
    """
    Abstract serializer used for verifying and refreshing JWTs.
    """
    token = serializers.CharField()

    def validate(self, attrs):
        msg = 'Please define a validate method.'
        raise NotImplementedError(msg)

    def _check_payload(self, token):
        # Check payload valid (based off of JSONWebTokenAuthentication,
        # may want to refactor)
        try:
            payload = jwt_decode_handler(token)
        except jwt.ExpiredSignature:
            msg = _('Signature has expired.')
            raise serializers.ValidationError(msg)
        except jwt.DecodeError:
            msg = _('Error decoding signature.')
            raise serializers.ValidationError(msg)

        return payload

    def _check_user(self, payload):
        username = jwt_get_username_from_payload(payload)

        if not username:
            msg = _('Invalid payload.')
            raise serializers.ValidationError(msg)

        # Make sure user exists
        try:
            user = User.objects.get_by_natural_key(username)
        except User.DoesNotExist:
            msg = _("User doesn't exist.")
            raise serializers.ValidationError(msg)

        if not user.is_active:
            msg = _('User account is disabled.')
            raise serializers.ValidationError(msg)

        return user


class VerifyJSONWebTokenSerializer(VerificationBaseSerializer):
    """
    Check the veracity of an access token.
    """

    # private = False

    def post_task_after_token_valid(self, token_obj):
        # set new ip address
        try:
            new_ip = get_client_ip(self.context.get('request'))
            token_obj.ip = new_ip
        except:
            pass

        token_obj.save()

    def validate(self, attrs):
        token = attrs['token']

        payload = self._check_payload(token=token)
        user = self._check_user(payload=payload)

        lookup_id = payload.get('lookup_id')
        user_id = user.id
        token_obj = TokenLookUpID.objects.filter(user__id=user_id, id=lookup_id).first()
        if token_obj:
            self.post_task_after_token_valid(token_obj)

            return {
                'token': token,
                'user': user
            }
        # raise ValidationError('توکن دیگر معتبر نیست.')
        raise ValidationError('Token is not valid anymore!')


class VerifyJSONWebTokenPrivateSerializer(VerifyJSONWebTokenSerializer):

    def post_task_after_token_valid(self, token_obj):
        pass

    # private = True


class LookUpIDSerializer(serializers.ModelSerializer):
    class Meta:
        model = TokenLookUpID
        fields = '__all__'

    def to_representation(self, instance):
        r = super(LookUpIDSerializer, self).to_representation(instance)
        current = False
        request = self.context.get('request')

        # Find out which token user is using.
        lookup_id = get_lookup_id_from_request(request)
        if lookup_id == instance.id:
            current = True

        r.update({
            'current': current
        })
        return r


class DeleteListOfTokenSerializer(serializers.Serializer):
    id_list = serializers.PrimaryKeyRelatedField(queryset=TokenLookUpID.objects.all(), many=True,
                                                 required=True, allow_empty=False)



# Generated by Django 2.2.5 on 2019-10-18 13:25

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('token_manager', '0001_initial'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AddField(
            model_name='tokenlookupid',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
    ]



import jwt

from django.contrib.auth import get_user_model
from django.utils.encoding import smart_text
from django.utils.translation import ugettext as _
from rest_framework import exceptions
from rest_framework.authentication import (
    BaseAuthentication, get_authorization_header
)

from .settings import api_settings
from .models import TokenLookUpID

jwt_decode_handler = api_settings.JWT_DECODE_HANDLER
jwt_get_username_from_payload = api_settings.JWT_PAYLOAD_GET_USERNAME_HANDLER


class FakeAuthentication(BaseAuthentication):

    def authenticate(self, request):
        User = get_user_model()
        user, _ = User.objects.get_or_create(username='test')
        if not user.is_superuser:
            user.is_superuser = True
            user.is_staff = True
            user.save()
        return (user, None)


class BaseJSONWebTokenAuthentication(BaseAuthentication):
    """
    Token based authentication using the JSON Web Token standard.
    """

    def authenticate(self, request):
        """
        Returns a two-tuple of `User` and token if a valid signature has been
        supplied using JWT-based authentication.  Otherwise returns `None`.
        """
        jwt_value = self.get_jwt_value(request)
        if jwt_value is None:
            return None

        try:
            payload = jwt_decode_handler(jwt_value)
        except jwt.ExpiredSignature:
            msg = _('Signature has expired.')
            raise exceptions.AuthenticationFailed(msg)
        except jwt.DecodeError:
            msg = _('Error decoding signature.')
            raise exceptions.AuthenticationFailed(msg)
        except jwt.InvalidTokenError:
            raise exceptions.AuthenticationFailed()

        user = self.authenticate_credentials(payload)

        return (user, jwt_value)

    def authenticate_credentials(self, payload):
        """
        Returns an active user that matches the payload's user id and email.
        """

        # First check if this token is available in database and is assigned to any user.
        TokenLookUpID.objects.check_token(user_id=payload.get('user_id', 0), lookup_id=payload.get('lookup_id', 0))

        User = get_user_model()
        username = jwt_get_username_from_payload(payload)

        if not username:
            msg = _('Invalid payload.')
            raise exceptions.AuthenticationFailed(msg)

        try:
            user = User.objects.get_by_natural_key(username)
        except User.DoesNotExist:
            msg = _('Invalid signature.')
            raise exceptions.AuthenticationFailed(msg)

        if not user.is_active:
            msg = _('User account is disabled.')
            raise exceptions.AuthenticationFailed(msg)

        return user


class JSONWebTokenAuthentication(BaseJSONWebTokenAuthentication):
    """
    Clients should authenticate by passing the token key in the "Authorization"
    HTTP header, prepended with the string specified in the setting
    `JWT_AUTH_HEADER_PREFIX`. For example:

        Authorization: JWT eyJhbGciOiAiSFMyNTYiLCAidHlwIj
    """
    www_authenticate_realm = 'api'

    def get_jwt_value(self, request):
        auth = get_authorization_header(request).split()
        auth_header_prefix = api_settings.JWT_AUTH_HEADER_PREFIX.lower()

        if not auth:
            if api_settings.JWT_AUTH_COOKIE:
                return request.COOKIES.get(api_settings.JWT_AUTH_COOKIE)
            return None

        if smart_text(auth[0].lower()) != auth_header_prefix:
            return None

        if len(auth) == 1:
            msg = _('Invalid Authorization header. No credentials provided.')
            raise exceptions.AuthenticationFailed(msg)
        elif len(auth) > 2:
            msg = _('Invalid Authorization header. Credentials string '
                    'should not contain spaces.')
            raise exceptions.AuthenticationFailed(msg)

        return auth[1]

    def authenticate_header(self, request):
        """
        Return a string to be used as the value of the `WWW-Authenticate`
        header in a `401 Unauthenticated` response, or `None` if the
        authentication scheme should return `403 Permission Denied` responses.
        """
        return '{0} realm="{1}"'.format(api_settings.JWT_AUTH_HEADER_PREFIX, self.www_authenticate_realm)



from django.contrib.auth import get_user_model

from rest_framework import serializers


class Serializer(serializers.Serializer):
    @property
    def object(self):
        return self.validated_data


class PasswordField(serializers.CharField):

    def __init__(self, *args, **kwargs):
        if 'style' not in kwargs:
            kwargs['style'] = {'input_type': 'password'}
        else:
            kwargs['style']['input_type'] = 'password'
        super(PasswordField, self).__init__(*args, **kwargs)


def get_username_field():
    try:
        username_field = get_user_model().USERNAME_FIELD
    except:
        username_field = 'username'

    return username_field


def get_username(user):
    try:
        username = user.get_username()
    except AttributeError:
        username = user.username

    return username



# Generated by Django 2.2.5 on 2019-10-18 13:25

from django.db import migrations, models
import django_jalali.db.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='TokenLookUpID',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('ip', models.CharField(blank=True, max_length=20, null=True)),
                ('os', models.CharField(blank=True, max_length=300, null=True)),
                ('r_type', models.CharField(blank=True, max_length=20, null=True)),
                ('device', models.CharField(blank=True, max_length=20, null=True)),
                ('browser', models.CharField(blank=True, max_length=30, null=True)),
                ('created_at', django_jalali.db.models.jDateTimeField(auto_now_add=True)),
                ('updated_at', django_jalali.db.models.jDateTimeField(auto_now=True)),
            ],
            options={
                'db_table': 'token_manager_lookup_id',
            },
        ),
    ]



import os
from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.md')).read()

setup(
    name='django-token-manager',
    version='1.1.1',
    packages=['token_manager'],
    description='A Django Rest Token Authentication system like telegram which will be using JWT as core with extended features.',
    long_description=README,
    long_description_content_type='text/markdown',
    author='Execut3',
    author_email='execut3.binarycodes@gmail.com',
    url='https://github.com/Execut3/django-token-manager',
    license='GPT',
    install_requires=[
        'Django>=2.0',
        'djangorestframework>=3.0',
        'django-jalali>=3.1.0',
        'django-rest-captcha==0.1.0',
        'django-user-agents>=0.3.2',
    ]
)



from datetime import datetime

from rest_framework import status, mixins
from rest_framework.decorators import action
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_framework.viewsets import GenericViewSet

from .models import TokenLookUpID
from .serializers import (
    JSONWebTokenSerializer, VerifyJSONWebTokenSerializer, LookUpIDSerializer,
    DeleteListOfTokenSerializer, JSONWebTokenWithCaptchaSerializer)
from .settings import api_settings
from .utils import jwt_decode_handler, get_lookup_id_from_request

jwt_response_payload_handler = api_settings.JWT_RESPONSE_PAYLOAD_HANDLER


class JSONWebTokenAPIView(APIView):
    """
    Base API View that various JWT interactions inherit from.
    """
    permission_classes = ()
    authentication_classes = ()

    def get_serializer_context(self):
        """
        Extra context provided to the serializer class.
        """
        return {
            'request': self.request,
            'view': self,
        }

    def get_serializer_class(self):
        """
        Return the class to use for the serializer.
        Defaults to using `self.serializer_class`.
        You may want to override this if you need to provide different
        serializations depending on the incoming request.
        (Eg. admins get full serialization, others get basic serialization)
        """
        assert self.serializer_class is not None, (
                "'%s' should either include a `serializer_class` attribute, "
                "or override the `get_serializer_class()` method."
                % self.__class__.__name__)
        return self.serializer_class

    def get_serializer(self, *args, **kwargs):
        """
        Return the serializer instance that should be used for validating and
        deserializing input, and for serializing output.
        """
        serializer_class = self.get_serializer_class()
        kwargs['context'] = self.get_serializer_context()
        return serializer_class(*args, **kwargs)

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)

        if serializer.is_valid():
            user = serializer.object.get('user') or request.user
            token = serializer.object.get('token')

            response_data = jwt_response_payload_handler(token, user, request)

            # get lookup_id from payload
            # payload = jwt_decode_handler(token)
            # lookup_id = payload.get('lookup_id')
            # token_lookup = TokenLookUpID.objects.filter(user__id=user.id, id=lookup_id).first()
            # if token_lookup:
            #     response_data.update({
            #         'extra': {
            #             'r_type': token_lookup.r_type, 'ip': token_lookup.ip,
            #             'device': token_lookup.device, 'os': token_lookup.os,
            #             'browser': token_lookup.browser, 'created_at': str(token_lookup.created_at),
            #         }
            #     })

            response = Response(response_data)
            if api_settings.JWT_AUTH_COOKIE:
                expiration = (datetime.utcnow() +
                              api_settings.JWT_EXPIRATION_DELTA)
                response.set_cookie(api_settings.JWT_AUTH_COOKIE,
                                    token,
                                    expires=expiration,
                                    httponly=True)
            return response

        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)


class ObtainJSONWebToken(JSONWebTokenAPIView):
    """
    API View that receives a POST with a user's username and password.

    Returns a JSON Web Token that can be used for authenticated requests.
    """
    serializer_class = JSONWebTokenSerializer


class ObtainJSONWebTokenAdmin(ObtainJSONWebToken):

    serializer_class = JSONWebTokenWithCaptchaSerializer


class VerifyJSONWebToken(JSONWebTokenAPIView):
    """
    API View that checks the veracity of a token, returning the token if it
    is valid.
    """
    serializer_class = VerifyJSONWebTokenSerializer


class JSONWebTokenView(mixins.RetrieveModelMixin,
                       mixins.DestroyModelMixin,
                       mixins.ListModelMixin,
                       GenericViewSet):
    serializer_class = LookUpIDSerializer
    pagination_class = None

    def get_queryset(self):
        request_user = self.request.user
        queryset = TokenLookUpID.objects.all()

        if not request_user.is_staff:
            queryset = TokenLookUpID.objects.filter(user__id=request_user.id)

        user_id = self.request.query_params.get('user_id', '')
        if user_id:
            queryset = queryset.filter(user__id=user_id)
        return queryset

    def get_object(self):
        obj = super(JSONWebTokenView, self).get_object()
        if self.request.user == obj.user or self.request.user.is_staff:
            return obj
        # raise ValidationError('دسترسی لازم به این توکن را ندارید.')
        raise ValidationError('You don\'t have access to this token!')

    @action(detail=False, methods=['post'], url_path='delete-list', serializer_class=DeleteListOfTokenSerializer)
    def delete_tokens_by_list(self, request, **kwargs):

        serializer = self.get_serializer(data=request.data)

        if serializer.is_valid():

            tokens = serializer.validated_data.get('id_list')

            # Remove current token from list if exist (Should not be able to remove token that is sending this request.)
            current_token_id = get_lookup_id_from_request(request)
            current_token = TokenLookUpID.objects.filter(id=current_token_id).first()
            if current_token in tokens:
                tokens.remove(current_token)

            token_users = []
            for t in tokens:
                if not t.user in token_users:
                    token_users.append(t.user)

            # If not is_staff, can only access tokens of his own.
            if not self.request.user.is_staff:
                for u in token_users:
                    if not self.request.user == u:
                        # raise ValidationError('شما دسترسی لازم برای حذف این توکن‌ها را ندارید.')
                        raise ValidationError('You don\'t have required permissions to do this operation!')

            count = len(tokens)
            for token in tokens:
                token.delete()

            # return Response('تعداد {} توکن با موفقیت حذف شدند.'.format(count), status=status.HTTP_204_NO_CONTENT)
            return Response('Total {} tokens removed successfully.'.format(count), status=status.HTTP_204_NO_CONTENT)

        return Response(serializer.errors, status=status.HTTP_404_NOT_FOUND)


class RemoveTokenView(APIView):
    """
    Api view to logout user and remove it's token
    """

    def get(self, request, *args, **kwargs):
        try:
            lookup_id = get_lookup_id_from_request(request)
            token = TokenLookUpID.objects.get(id=lookup_id)
            token.delete()
        except Exception as e:
            print(e)
            # raise ValidationError('کاربر وارد نشده است')
            raise ValidationError('User is not login!')
        # return Response({'status_code': 200, 'message': 'کاربر با موفقیت خارج شد.'})
        return Response({'status_code': 200, 'message': 'User successfully exited.'})

    def post(self, request, *args, **kwargs):
        try:
            lookup_id = get_lookup_id_from_request(request)
            token = TokenLookUpID.objects.get(id=lookup_id)
            token.delete()
        except Exception as e:
            print(e)
            # raise ValidationError('کاربر وارد نشده است')
            raise ValidationError('User is not login!')
        # return Response({'status_code': 200, 'message': 'کاربر با موفقیت خارج شد.'})
        return Response({'status_code': 200, 'message': 'User successfully exited.'})


logout_view = RemoveTokenView.as_view()


obtain_jwt_token = ObtainJSONWebToken.as_view()
obtain_jwt_token_admin = ObtainJSONWebTokenAdmin.as_view()
verify_jwt_token = VerifyJSONWebToken.as_view()



import jwt
import uuid
import warnings

from django.contrib.auth import get_user_model

from calendar import timegm
from datetime import datetime

from rest_framework_jwt.compat import get_username
from rest_framework_jwt.compat import get_username_field
from rest_framework_jwt.settings import api_settings

from .models import TokenLookUpID

User = get_user_model()


def jwt_get_secret_key(payload=None):
    """
    For enhanced security you may want to use a secret key based on user.

    This way you have an option to logout only this user if:
        - token is compromised
        - password is changed
        - etc.
    """
    if api_settings.JWT_GET_USER_SECRET_KEY:
        User = get_user_model()  # noqa: N806
        user = User.objects.get(pk=payload.get('user_id'))
        key = str(api_settings.JWT_GET_USER_SECRET_KEY(user))
        return key
    return api_settings.JWT_SECRET_KEY


def jwt_payload_handler(user, ip, r_type, browser, os, device):
    username_field = get_username_field()
    username = get_username(user)

    warnings.warn(
        'The following fields will be removed in the future: '
        '`email` and `user_id`. ',
        DeprecationWarning
    )

    # Create a lookupID for this user, also check if it exceeded max allowed or other situations...
    lookup_id = TokenLookUpID.objects.create_token(user, ip, r_type, browser, os, device)

    try:
        user.last_login = datetime.now()
        user.save()
    except Exception as e:
        print(e)

    # Update payload with new lookup_id, it is needed for verifications of token. Because when we revoke a token,
    # this lookup_id will not be available for this user anymore.
    payload = {
        'lookup_id': lookup_id.id,
        'user_id': user.pk,
        'username': username,
        'exp': datetime.utcnow() + api_settings.JWT_EXPIRATION_DELTA
    }
    if hasattr(user, 'email'):
        payload['email'] = user.email
    if isinstance(user.pk, uuid.UUID):
        payload['user_id'] = str(user.pk)

    payload[username_field] = username

    # Include original issued at time for a brand new token,
    # to allow token refresh
    if api_settings.JWT_ALLOW_REFRESH:
        payload['orig_iat'] = timegm(
            datetime.utcnow().utctimetuple()
        )

    if api_settings.JWT_AUDIENCE is not None:
        payload['aud'] = api_settings.JWT_AUDIENCE

    if api_settings.JWT_ISSUER is not None:
        payload['iss'] = api_settings.JWT_ISSUER

    return payload, lookup_id


def jwt_get_user_id_from_payload_handler(payload):
    """
    Override this function if user_id is formatted differently in payload
    """
    warnings.warn(
        'The following will be removed in the future. '
        'Use `JWT_PAYLOAD_GET_USERNAME_HANDLER` instead.',
        DeprecationWarning
    )

    return payload.get('user_id')


def jwt_get_username_from_payload_handler(payload):
    """
    Override this function if username is formatted differently in payload
    """
    return payload.get('username')


def jwt_encode_handler(payload):
    key = api_settings.JWT_PRIVATE_KEY or jwt_get_secret_key(payload)
    return jwt.encode(
        payload,
        key,
        api_settings.JWT_ALGORITHM
    ).decode('utf-8')


def jwt_decode_handler(token):
    options = {
        'verify_exp': api_settings.JWT_VERIFY_EXPIRATION,
    }
    # get user from token, BEFORE verification, to get user secret key
    unverified_payload = jwt.decode(token, None, False)
    secret_key = jwt_get_secret_key(unverified_payload)
    return jwt.decode(
        token,
        api_settings.JWT_PUBLIC_KEY or secret_key,
        api_settings.JWT_VERIFY,
        options=options,
        leeway=api_settings.JWT_LEEWAY,
        audience=api_settings.JWT_AUDIENCE,
        issuer=api_settings.JWT_ISSUER,
        algorithms=[api_settings.JWT_ALGORITHM]
    )


def jwt_response_payload_handler(token, user=None, request=None):
    """
    Returns the response data for both the login and refresh views.
    Override to return a custom response such as including the
    serialized representation of the User.

    Example:

    def jwt_response_payload_handler(token, user=None, request=None):
        return {
            'token': token,
            'user': UserSerializer(user, context={'request': request}).data
        }

    """
    # from permission.utils import get_user_permissions
    # from user_management.serializers import UserSerializer
    # user_data = UserSerializer().to_representation(user)
    # user_data['balance'] = user.wallet.balance
    return {
        'token': token,
        # 'user': user_data,
        # 'permissions': get_user_permissions(user)
    }


def get_lookup_id_from_request(request):
    auth_token = getattr(request, 'auth', None)
    lookup_id = None
    if auth_token:
        decoded_jwt = jwt_decode_handler(auth_token)
        lookup_id = decoded_jwt.get('lookup_id', None)
    return lookup_id


def delete_all_user_tokens(user_id):
    TokenLookUpID.objects.filter(user__id=user_id).delete()


def fetch_user_from_token(token):
    """
    This method simply, just retreive token as an argument, then tries to validate token and if the token is available,
    Then will return the corresponded user of it.
    :param token:
    :return:
    """
    try:
        decoded_jwt = jwt_decode_handler(token)
        user_id = decoded_jwt['user_id']
        lookup_id = decoded_jwt['lookup_id']
        token_obj = TokenLookUpID.objects.filter(user__id=user_id, id=lookup_id).first()
        if token_obj:
            return User.objects.get(id=user_id)
    except Exception as e:
        print(e)
        return None
    finally:
        return None


def get_client_ip(request):
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR')
    return ip


def fetch_user_agent_info(user_agent):
    if type(user_agent) == str:
        r_type = user_agent
        browser = 'Other'
        os = user_agent
        device = user_agent
        return r_type, browser, os, device

    try:
        if user_agent.is_mobile:
            r_type = 'mobile'
        elif user_agent.is_tablet:
            r_type = 'tablet'
        elif user_agent.is_touch_capable:
            r_type = 'touch_capable'
        elif user_agent.is_pc:
            r_type = 'pc'
        elif user_agent.is_bot:
            r_type = 'bot'
        else:
            r_type = 'other'
    except:
        r_type = 'other'

    # Accessing user agent's browser attributes
    try:
        browser = '{} {}'.format(user_agent.browser.family, user_agent.browser.version_string)
    except:
        browser = 'Other'

    # Operating System properties
    try:
        os = '{} {}'.format(user_agent.os.family, user_agent.os.version_string)
    except:
        os = 'Other'

    # Device properties
    try:
        device = user_agent.device.family
    except:
        device = 'Other'

    return r_type, browser, os, device


def fetch_request_extra_info(request):
    custom_useragent = request.META.get('HTTP_USERAGENT', '')
    custom_machinehost = request.META.get('HTTP_MACHINE_HOST_NAME', '')
    if custom_machinehost or custom_useragent:
        r_type, browser, os, device = custom_useragent, custom_useragent, custom_useragent, custom_machinehost
    else:
        user_agent = request.user_agent
        r_type, browser, os, device = fetch_user_agent_info(user_agent)
    ip = get_client_ip(request)
    return ip, r_type, browser, os, device


def get_token_from_request(request, user):
    ip, r_type, browser, os, device = fetch_request_extra_info(request)
    payload, _ = jwt_payload_handler(user, ip, r_type, browser, os, device)
    return jwt_encode_handler(payload)
